@@ -4,7 +4,7 @@
 #
 # timsieved client script
 #
-# Copyright © 2006-2010 Phil Pennock.  All rights reserved.
+# Copyright © 2006-2011 Phil Pennock.  All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions
@@ -33,6 +33,12 @@
 use warnings;
 use strict;
 
+# If you can't update /etc/services to contain an entry for 'sieve' and you're
+# not using 4190 (specified in RFC 5804) and you're not publishing an SRV
+# record, then you might want to change the default port-number in the
+# parentheses here:
+my $DEFAULT_PORT = 'sieve(4190)';
+
 # These are the defaults, some may be overriden on the command-line.
 my %ssl_options = (
 	SSL_version	=> 'TLSv1',
@@ -49,7 +55,7 @@ my @cmd_localfs_ls = qw( ls -C );
 # No user-serviceable parts below
 
 our $VERSION = 0;
-my $svn_revision = '$Revision: 72 $';
+my $svn_revision = '$Revision: 80 $';
 if ($svn_revision =~ /^.Revision:\s*(\d+)\s*\$\z/) {
 	$svn_revision = $1;
 	$VERSION = '0.' . $1;
@@ -75,6 +81,7 @@ use IO::File;
 use IO::Socket::INET6;
 use IO::Socket::SSL 0.97; # SSL_ca_path bogus before 0.97
 use MIME::Base64;
+use Net::DNS;
 use Pod::Usage;
 use POSIX qw/ strftime /;
 use Term::ReadKey;
@@ -98,6 +105,7 @@ sub do_version_display {
 				'Authen::SASL::Perl',
 				'IO::Socket::INET6',
 				'IO::Socket::SSL',
+				'Net::DNS',
 				'Term::ReadKey',
 				@do_require) {
 			my $vname = "${mod}::VERSION";
@@ -127,12 +135,14 @@ my $DEBUGGING_SASL = 0;
 my $DATASTART = tell DATA;
 my $localsievename;
 my $remotesievename;
+my $port = undef;
 my ($user, $authzid, $authmech, $sslkeyfile, $sslcertfile, $passwordfd);
 my $prioritise_auth_external = 0;
 my $dump_tls_information = 0;
 my $opt_version_req = 0;
+my $ignore_server_version = 0;
+my $no_srv = 0;
 my ($server, $realm);
-my $port = 'sieve(2000)';
 my $net_domain = AF_UNSPEC;
 my $action = 'command-loop';
 my $execscript;
@@ -141,6 +151,7 @@ GetOptions(
 	"remotesieve=s"	=> \$remotesievename,
 	"server|s=s"	=> \$server,
 	"port|p=s"	=> \$port, # not num, allow service names
+	"nosrv"		=> \$no_srv,
 	"user|u=s"	=> \$user,
 	"realm|r=s"	=> \$realm,
 	"authzid|authname|a=s"	=> \$authzid, # authname for sieveshell compat
@@ -157,6 +168,7 @@ GetOptions(
 	"debug"		=> \$DEBUGGING,
 	"debugsasl"	=> \$DEBUGGING_SASL,
 	"dumptlsinfo|dumpsslinfo"	=> \$dump_tls_information,
+	"ignoreserverversion"	=> \$ignore_server_version,
 	# option names can be short-circuited, $action is complete:
 	"upload"	=> sub { $action = 'upload' },
 	"download"	=> sub { $action = 'download' },
@@ -164,6 +176,7 @@ GetOptions(
 	"delete"	=> sub { $action = 'delete' },
 	"activate"	=> sub { $action = 'activate' },
 	"deactivate"	=> sub { $action = 'deactivate' },
+	"checkscript"   => sub { $action = 'checkscript' },
 	"exec|e=s"	=> sub { $execscript = $_[1]; $action='command-loop' },
 	'help|?'	=> sub { pod2usage(0) },
 	'man'		=> sub { pod2usage(-exitstatus => 0, -verbose => 2) },
@@ -199,9 +212,11 @@ unless (defined $server) {
 }
 
 die "Bad server name\n"
-	unless $server =~ /^[A-Za-z0-9_.-]+\z/;
-die "Bad port specification\n"
-	unless $port =~ /^[A-Za-z0-9_()-]+\z/;
+	unless $server =~ /^[A-Za-z0-9_.:-]+\z/;
+if (defined $port) {
+	die "Bad port specification\n"
+		unless $port =~ /^[A-Za-z0-9_()-]+\z/;
+}
 
 unless (defined $user) {
 	if ($^O eq "MSWin32") {
@@ -237,7 +252,7 @@ if (defined $localsievename and not defined $remotesievename) {
 	$remotesievename = $localsievename;
 }
 
-if (defined $localsievename and $action eq 'upload') {
+if (defined $localsievename and $action =~ /upload|checkscript/) {
 	-r $localsievename or die "unable to read \"$localsievename\": $!\n";
 }
 if ($action eq 'download' and not defined $localsievename) {
@@ -257,23 +272,73 @@ if ($action eq 'deactivate' and defined $remotesievename) {
 # ######################################################################
 # Start work; connect, start TLS, authenticate
 
-my $sock = IO::Socket::INET6->new(
-	PeerHost	=> $server,
-	PeerPort	=> $port,
-	Proto		=> 'tcp',
-	Domain		=> $net_domain,
-	MultiHomed	=> 1, # try multiple IPs (IPv4 works, v6 doesn't?)
-);
-unless (defined $sock) {
-	my $extra = '';
-	if ($!{EINVAL} and $net_domain != AF_UNSPEC) {
-	  $extra = " (Probably no host record for overriden IP version)\n";
+my @host_port_pairs;
+
+# Find the real hostname to connect to.
+unless ($no_srv) {
+	my $res = Net::DNS::Resolver->new();
+	my $query = $res->query("_sieve._tcp.$server", 'SRV');
+	my @srv_recs;
+	if ($query) {
+		foreach my $rr ($query->answer) {
+			next unless $rr->type eq 'SRV';
+			push @srv_recs, $rr;
+		}
 	}
-	die qq{Connection to "$server" [port $port] failed: $!\n$extra};
+	if (@srv_recs) {
+		@srv_recs = Net::DNS::rrsort('SRV', '', @srv_recs);
+		my $old = "[$server]:" . (defined $port ? $port : $DEFAULT_PORT);
+		debug "dns: SRV results found for: $old";
+		foreach my $rr (@srv_recs) {
+			push @host_port_pairs, [$rr->target, $rr->port];
+		}
+	}
+
 }
 
+$port = $DEFAULT_PORT unless defined $port;
+unless (@host_port_pairs) {
+	push @host_port_pairs, [$server, $port];
+}
+my $sock = undef;
+
+# Yes, this used to just try one connection and the list of candidates was
+# bolted on; how could you tell?
+
+foreach my $hp (@host_port_pairs) {
+	my $host_candidate = $hp->[0];
+	my $port_candidate = $hp->[1];
+	debug "connection: trying <[$host_candidate]:$port_candidate>";
+	my $s = IO::Socket::INET6->new(
+		PeerHost	=> $host_candidate,
+		PeerPort	=> $port_candidate,
+		Proto		=> 'tcp',
+		Domain		=> $net_domain,
+		MultiHomed	=> 1, # try multiple IPs (IPv4 works, v6 doesn't?)
+	);
+	unless (defined $s) {
+		my $extra = '';
+		if ($!{EINVAL} and $net_domain != AF_UNSPEC) {
+		  $extra = " (Probably no host record for overriden IP version)\n";
+		}
+		warn qq{Connection to <[$host_candidate]:$port_candidate> failed: $!\n$extra};
+		next;
+	}
+	unless ($s->peerhost()) {
+		# why am I seeing successful returns for unconnected sockets? *sigh*
+		warn qq{Connection to <[$host_candidate]:$port_candidate> failed.\n};
+		next;
+	}
+	$sock = $s;
+	$server = $host_candidate;
+	$port = $port_candidate;
+	last;
+}
+exit(1) unless defined $sock;
+
 $sock->autoflush(1);
-debug "connection: remote host address is @{[$sock->peerhost()]}";
+debug "connection: remote host address is [@{[$sock->peerhost()]}] " .
+	"port [@{[$sock->peerport()]}]";
 
 my %capa;
 my %raw_capabilities;
@@ -435,7 +500,7 @@ my $tls_bitlength = -1;
 if (exists $capa{STARTTLS}) {
 	ssend $sock, "STARTTLS";
 	sget $sock;
-	die "STARTTLS request rejected: $_\n" unless /^OK\s+\"/;
+	die "STARTTLS request rejected: $_\n" unless /^OK\b/;
 	IO::Socket::SSL->start_SSL($sock, %ssl_options) or do {
 		my $e = IO::Socket::SSL::errstr();
 		die "STARTTLS promotion failed: $e\n";
@@ -632,7 +697,7 @@ if (defined $realm) {
 	}
 
 	if (/^NO((?:\s.*)?)$/) {
-		closedie_NOmsg($sock, $1, "Authentication refused by server");
+		closedie_NOmsg($sock, $1, "Authentication refused by server\n");
 	}
 	if (/^OK\s+\(SASL\s+\"([^"]+)\"\)$/) {
 		# This _should_ be present with server-verification steps which
@@ -659,7 +724,7 @@ if (defined $realm) {
 			}
 		}
 		if (defined $valid and length $valid) {
-			closedie($sock, "Server failed final verification [$valid]");
+			closedie($sock, "Server failed final verification [$valid]\n");
 		}
 	}
 
@@ -674,6 +739,7 @@ sub sieve_activate;
 sub sieve_delete;
 sub sieve_download;
 sub sieve_upload;
+sub sieve_checkscript;
 sub localfs_ls;
 sub localfs_chpwd;
 sub localfs_pwd;
@@ -698,6 +764,8 @@ sub tilde_expand ; # don't apply to cmdline params as shell does it for us
 # param list numbering: 1=first param, ...
 # 'remote_name' => if there's a remote name, which position it comes
 # 'local_name' => if there's a local name, which position it comes
+# 'min_version' => require server to advertise this version for support
+#                  (undef => "advertises VERSION capability")
 my %sieve_commands = (
 	help	=> { routine => \&aux_help, params => 0, help => 'this help' },
 	'?'	=> { alias => 'help' },
@@ -763,6 +831,14 @@ my %sieve_commands = (
 		local_name => 1,
 		remote_name => 2,
 	},
+	checkscript  => {
+		routine => \&sieve_checkscript,
+		help => '<filename> -- check script on the server',
+		action => 1,
+		params => 1,
+		local_name => 1,
+		min_version => undef,
+	},
 	put	=> { alias => 'upload' },
 	download => {
 		routine => \&sieve_download,
@@ -826,6 +902,42 @@ unless ($have_needed_man_mods) {
 	delete $sieve_commands{'man'};
 }
 
+# ######################################################################
+# Fix-up for features missing in this server
+
+unless ($ignore_server_version) {
+	# If server does not advertise VERSION, it's missing certain features;
+	# if it does, we can do min_version checks.
+	# If our min_version is undef, we simply require any VERSION
+	# RFC5804 defines VERSION as just a string "1.0" and says nothing
+	# about comparison.  So for the time being, we'll go on "must look like
+	# a number, optionally with a dot in it, and compares with Perl's
+	# numerical operator" as a good-enough approach to predict the future.
+	my $have_server_version = undef;
+	if (exists $capa{VERSION}) {
+		closedie($sock, "Unparsed server version [$capa{VERSION}]\n")
+			unless $capa{VERSION} =~ /^[0-9]+(?:\.[0-9]+)?\z/;
+		$have_server_version = $capa{VERSION};
+	};
+
+	my @kl = keys %sieve_commands;
+	foreach my $k (@kl) {
+		next unless exists $sieve_commands{$k}{min_version};
+		my $min = $sieve_commands{$k}{min_version};
+		unless (defined $have_server_version) {
+			delete $sieve_commands{$k};
+			next;
+		}
+		unless (defined $min) {
+			next;
+		}
+		if ($min > $have_server_version) {
+			delete $sieve_commands{$k};
+			next;
+		}
+	}
+}
+
 # ######################################################################
 # Do something
 
@@ -1162,6 +1274,33 @@ sub sieve_upload
 	}
 }
 
+sub sieve_checkscript
+{
+	my ($sock, $localfn) = @_; splice @_, 0, 2;
+	die "QUIT:Internal error, check missing localfn\n"
+	    unless defined $localfn;
+
+	# I'm going to assume that any Sieve script will easily fit in memory.
+	# Since Cyrus enforces admin-specified size constraints, this is
+	# probably pretty safe.
+	my $fh = new IO::File tilde_expand($localfn), '<'
+		or die "aborting, read-open($localfn) failed: $!\n";
+	my @scriptlines = $fh->getlines();
+	$fh->close() or die "aborting, read-close($localfn failed: $!\n";
+
+	my $len = 0;
+	$len += length foreach @scriptlines;
+
+	ssend $sock, "CHECKSCRIPT {${len}+}";
+	ssend $sock, '-noeol', @scriptlines;
+	ssend $sock, '';
+	sget $sock;
+
+	unless (/^OK((?:\s.*)?)$/) {
+		warn "CHECKSCRIPT failed: $_\n";
+	}
+}
+
 sub localfs_ls
 {
 	my ($sock, $localdir) = @_;
@@ -1516,14 +1655,15 @@ sieve-connect - managesieve command-line client
 	       [--notlsverify|--nosslverify]
 	       [--noclearauth] [--noclearchan]
 	       [--authmech <mechanism>]
-	       [--upload|--download|--list|--delete|
+	       [--ignoreserverversion]
+	       [--upload|--download|--list|--delete|--check
 	        --activate|--deactivate]|[--exec <script>]
 	       [--help|--man]
 
 =head1 DESCRIPTION
 
 B<sieve-connect> is a client for the C<MANAGESIEVE> protocol, which is
-an Internet Draft protocol for manipulation of C<Sieve> scripts in a
+an RFC-specified protocol for manipulation of C<Sieve> scripts in a
 repository.
 More simply, B<sieve-connect> lets you control your mail-filtering
 rule files on a mail server.
@@ -1568,9 +1708,12 @@ The server can be a host or IP address, IPv4 or IPv6;
 the default is C<$IMAP_SERVER> from the environment (if it's not a
 unix-domain socket path) with any port specificaion stripped off,
 else F<localhost>.
-The port can be any Perl port specification, default is F<sieve(2000)>.
+The port can be any Perl port specification, default is F<sieve(4190)>.
 The B<--4> or B<--6> options may be used to coerce IPv4 or IPv6.
 
+By default, the server is taken to be a domain, for which SRV records are
+looked up; use B<--nosrv> to inhibit SRV record lookup.
+
 The B<--user> option will be required unless you're on a Unix system
 with getpwuid() available and your Cyrus account name matches your system
 account name.  B<--authmech> can be used to force a particular authentication
@@ -1600,6 +1743,9 @@ mechanisms unless protected by TLS.  The B<--noclearchan> option will
 mandate use of some confidentiality layer; at this time only TLS is
 supported.
 
+By default, the server's "VERSION" capability will be used to filter the
+commands available.  Use B<--ignoreserverversion> to prevent this.
+
 The remaining options denote actions.  One, and only one, action may be
 present.  If no action is present, the interactive mode is entered.
 If the exec action is present, commands are read from the script
@@ -1634,6 +1780,26 @@ If B<sieve-connect> fails to connect to an IPv4 server without the B<-4>
 option being explicitly passed, then you've encountered a portability
 issue in the F<IO::Socket::INET6> Perl library and need to upgrade that.
 
+Most historical implementations used port 2000 for ManageSieve.  RFC5804
+allocates port 4190.  This tool uses a port-spec of "sieve(4190)" as the
+default port, which means that an /etc/services (or substitute) entry for
+"sieve" as a TCP service takes precedence, but if that is not present, to
+assume 4190 as the default.  This change means that if you're still using
+port 2000 and do not have an /etc/services entry, updating to/beyond release
+0.75 of this tool will break invocations which do not specify a port.  The
+specification of the default port was moved to the user-configurable section
+at the top of the script and administrators may wish to override the shipped
+default.  You can bypass all of this mess by publishing SRV records,
+per RFC5804.
+
+The Net::DNS Perl module does not (at time of writing) provide full support for
+weighted prioritised SRV records and I have not made any effort to fix this;
+whatever the default sort algorithm provides for SRV is what is used for
+ordering.
+
+Probably need to sit down and work through the final RFC and see if any
+functionality is still missing.
+
 =head1 NON-BUGS
 
 Actually uses STARTTLS.  Can handle script names with embedded whitespace.
@@ -1651,10 +1817,16 @@ and deliberately designed to be compatible with sieveshell.
 
 Phil Pennock E<lt>phil-perl@spodhuis.orgE<gt> is guilty, m'Lud.
 
+There is a low-volume announcement list for new releases; the web interface is
+at L<http://mail.globnix.net/mailman/listinfo/sieve-connect-announce> or you
+can send mail,
+L<mailto:sieve-connect-announce-request@spodhuis.org?subject=subscribe>
+
 =head1 PREREQUISITES
 
 Perl.  F<Authen::SASL>.  F<IO::Socket::INET6>.
 F<IO::Socket::SSL> (at least version 0.97).  F<Pod::Usage>.
+F<Net::DNS> for SRV lookup.
 F<Pod::Simple::Text> for built-in man command (optional).
 F<Term::ReadKey> to get passwords without echo.
 Various other Perl modules which are believed to be standard.
